@@ -1,7 +1,9 @@
 /*
  *  linux/fs/ext2/truncate.c
  *
- *  Copyright (C) 1992, 1993  Remy Card (card@masi.ibp.fr)
+ *  Copyright (C) 1992, 1993, 1994  Remy Card (card@masi.ibp.fr)
+ *                                  Laboratoire MASI - Institut Blaise Pascal
+ *                                  Universite Pierre et Marie Curie (Paris VI)
  *
  *  from
  *
@@ -10,12 +12,30 @@
  *  Copyright (C) 1991, 1992  Linus Torvalds
  */
 
-#include <linux/sched.h>
-#include <linux/ext2_fs.h>
-#include <linux/tty.h>
-#include <linux/stat.h>
-#include <linux/fcntl.h>
+/*
+ * Real random numbers for secure rm added 94/02/18
+ * Idea from Pierre del Perugia <delperug@gla.ecoledoc.ibp.fr>
+ */
+
 #include <linux/errno.h>
+#include <linux/fs.h>
+#include <linux/ext2_fs.h>
+#include <linux/fcntl.h>
+#include <linux/sched.h>
+#include <linux/stat.h>
+#include <linux/locks.h>
+
+#define clear_block(addr,size,value) \
+	__asm__("cld\n\t" \
+		"rep\n\t" \
+		"stosl" \
+		: \
+		:"a" (value), "c" (size / 4), "D" ((long) (addr)) \
+		:"cx", "di")
+
+static int ext2_secrm_seed = 152;	/* Random generator base */
+
+#define RANDOM_INT (ext2_secrm_seed = ext2_secrm_seed * 69069l +1)
 
 /*
  * Truncate has the most races in the whole filesystem: coding it is
@@ -35,6 +55,8 @@ static int trunc_direct (struct inode * inode)
 	int i, tmp;
 	unsigned long * p;
 	struct buffer_head * bh;
+	unsigned long block_to_free = 0;
+	unsigned long free_count = 0;
 	int retry = 0;
 	int blocks = inode->i_sb->s_blocksize / 512;
 #define DIRECT_BLOCK ((inode->i_size + inode->i_sb->s_blocksize - 1) / \
@@ -47,7 +69,12 @@ repeat:
 		tmp = *p;
 		if (!tmp)
 			continue;
-		bh = get_hash_table (inode->i_dev, tmp, inode->i_sb->s_blocksize);
+		if (inode->u.ext2_i.i_flags & EXT2_SECRM_FL)
+			bh = getblk (inode->i_dev, tmp,
+				     inode->i_sb->s_blocksize);
+		else
+			bh = get_hash_table (inode->i_dev, tmp,
+					     inode->i_sb->s_blocksize);
 		if (i < direct_block) {
 			brelse (bh);
 			goto repeat;
@@ -60,9 +87,26 @@ repeat:
 		*p = 0;
 		inode->i_blocks -= blocks;
 		inode->i_dirt = 1;
+		if (inode->u.ext2_i.i_flags & EXT2_SECRM_FL) {
+			clear_block (bh->b_data, inode->i_sb->s_blocksize,
+				     RANDOM_INT);
+			bh->b_dirt = 1;
+		}
 		brelse (bh);
-		ext2_free_block (inode->i_sb, tmp);
+		if (free_count == 0) {
+			block_to_free = tmp;
+			free_count++;
+		} else if (free_count > 0 && block_to_free == tmp - free_count)
+			free_count++;
+		else {
+			ext2_free_blocks (inode->i_sb, block_to_free, free_count);
+			block_to_free = tmp;
+			free_count = 1;
+		}
+/*		ext2_free_blocks (inode->i_sb, tmp, 1); */
 	}
+	if (free_count > 0)
+		ext2_free_blocks (inode->i_sb, block_to_free, free_count);
 	return retry;
 }
 
@@ -72,6 +116,8 @@ static int trunc_indirect (struct inode * inode, int offset, unsigned long * p)
 	struct buffer_head * bh;
 	struct buffer_head * ind_bh;
 	unsigned long * ind;
+	unsigned long block_to_free = 0;
+	unsigned long free_count = 0;
 	int retry = 0;
 	int addr_per_block = EXT2_ADDR_PER_BLOCK(inode->i_sb);
 	int blocks = inode->i_sb->s_blocksize / 512;
@@ -100,8 +146,12 @@ repeat:
 		tmp = *ind;
 		if (!tmp)
 			continue;
-		bh = get_hash_table (inode->i_dev, tmp,
+		if (inode->u.ext2_i.i_flags & EXT2_SECRM_FL)
+			bh = getblk (inode->i_dev, tmp,
 				     inode->i_sb->s_blocksize);
+		else
+			bh = get_hash_table (inode->i_dev, tmp,
+					     inode->i_sb->s_blocksize);
 		if (i < indirect_block) {
 			brelse (bh);
 			goto repeat;
@@ -113,11 +163,28 @@ repeat:
 		}
 		*ind = 0;
 		ind_bh->b_dirt = 1;
+		if (inode->u.ext2_i.i_flags & EXT2_SECRM_FL) {
+			clear_block (bh->b_data, inode->i_sb->s_blocksize,
+				     RANDOM_INT);
+			bh->b_dirt = 1;
+		}
 		brelse (bh);
-		ext2_free_block (inode->i_sb, tmp);
+		if (free_count == 0) {
+			block_to_free = tmp;
+			free_count++;
+		} else if (free_count > 0 && block_to_free == tmp - free_count)
+			free_count++;
+		else {
+			ext2_free_blocks (inode->i_sb, block_to_free, free_count);
+			block_to_free = tmp;
+			free_count = 1;
+		}
+/*		ext2_free_blocks (inode->i_sb, tmp, 1); */
 		inode->i_blocks -= blocks;
 		inode->i_dirt = 1;
 	}
+	if (free_count > 0)
+		ext2_free_blocks (inode->i_sb, block_to_free, free_count);
 	ind = (unsigned long *) ind_bh->b_data;
 	for (i = 0; i < addr_per_block; i++)
 		if (*(ind++))
@@ -130,12 +197,16 @@ repeat:
 			*p = 0;
 			inode->i_blocks -= blocks;
 			inode->i_dirt = 1;
-			ext2_free_block (inode->i_sb, tmp);
+			ext2_free_blocks (inode->i_sb, tmp, 1);
 		}
+	if (IS_SYNC(inode) && ind_bh->b_dirt) {
+		ll_rw_block (WRITE, 1, &ind_bh);
+		wait_on_buffer (ind_bh);
+	}
 	brelse (ind_bh);
 	return retry;
 }
-		
+
 static int trunc_dindirect (struct inode * inode, int offset,
 			    unsigned long * p)
 {
@@ -186,8 +257,12 @@ repeat:
 			*p = 0;
 			inode->i_blocks -= blocks;
 			inode->i_dirt = 1;
-			ext2_free_block (inode->i_sb, tmp);
+			ext2_free_blocks (inode->i_sb, tmp, 1);
 		}
+	if (IS_SYNC(inode) && dind_bh->b_dirt) {
+		ll_rw_block (WRITE, 1, &dind_bh);
+		wait_on_buffer (dind_bh);
+	}
 	brelse (dind_bh);
 	return retry;
 }
@@ -225,8 +300,6 @@ repeat:
 			goto repeat;
 		tind = i + (unsigned long *) tind_bh->b_data;
 		retry |= trunc_dindirect(inode, EXT2_NDIR_BLOCKS +
-/*			addr_per_block + addr_per_block * addr_per_block +
-			(i * (addr_per_block * addr_per_block)), tind); */
 			addr_per_block + (i + 1) * addr_per_block * addr_per_block,
 			tind);
 		tind_bh->b_dirt = 1;
@@ -243,8 +316,12 @@ repeat:
 			*p = 0;
 			inode->i_blocks -= blocks;
 			inode->i_dirt = 1;
-			ext2_free_block (inode->i_sb, tmp);
+			ext2_free_blocks (inode->i_sb, tmp, 1);
 		}
+	if (IS_SYNC(inode) && tind_bh->b_dirt) {
+		ll_rw_block (WRITE, 1, &tind_bh);
+		wait_on_buffer (tind_bh);
+	}
 	brelse (tind_bh);
 	return retry;
 }
@@ -256,6 +333,7 @@ void ext2_truncate (struct inode * inode)
 	if (!(S_ISREG(inode->i_mode) || S_ISDIR(inode->i_mode) ||
 	    S_ISLNK(inode->i_mode)))
 		return;
+	ext2_discard_prealloc(inode);
 	while (1) {
 		retry = trunc_direct(inode);
 		retry |= trunc_indirect (inode, EXT2_IND_BLOCK,
@@ -266,19 +344,11 @@ void ext2_truncate (struct inode * inode)
 		retry |= trunc_tindirect (inode);
 		if (!retry)
 			break;
+		if (IS_SYNC(inode) && inode->i_dirt)
+			ext2_sync_inode (inode);
 		current->counter = 0;
 		schedule ();
 	}
 	inode->i_mtime = inode->i_ctime = CURRENT_TIME;
 	inode->i_dirt = 1;
 }
-
-/*
- * Called when a inode is released. Note that this is different
- * from ext2_open: open gets called at every open, but release
- * gets called only when /all/ the files are closed.
- */
-void ext2_release (struct inode * inode, struct file * filp)
-{
-	printk ("ext2_release not implemented\n");
-}
