@@ -14,6 +14,7 @@
 #include <linux/resource.h>
 #include <linux/mm.h>
 #include <linux/tty.h>
+#include <linux/malloc.h>
 
 #include <asm/segment.h>
 extern void shm_exit (void);
@@ -88,6 +89,8 @@ void release(struct task_struct * p)
 		if (task[i] == p) {
 			task[i] = NULL;
 			REMOVE_LINKS(p);
+			if (STACK_MAGIC != *(unsigned long *)p->kernel_stack_page)
+				printk(KERN_ALERT "release: %s kernel stack corruption. Aiee\n", p->comm);
 			free_page(p->kernel_stack_page);
 			free_page((long) p);
 			return;
@@ -347,7 +350,7 @@ static void forget_original_parent(struct task_struct * father)
 	}
 }
 
-volatile void do_exit(long code)
+NORET_TYPE void do_exit(long code)
 {
 	struct task_struct *p;
 	int i;
@@ -384,7 +387,7 @@ fake_volatile:
 	}
 
 	if (current->ldt) {
-		free_page((unsigned long) current->ldt);
+		vfree(current->ldt);
 		current->ldt = NULL;
 		for (i=1 ; i<NR_TASKS ; i++) {
 			if (task[i] == current) {
@@ -394,9 +397,6 @@ fake_volatile:
 		}
 	}
 
-	current->state = TASK_ZOMBIE;
-	current->exit_code = code;
-	current->rss = 0;
 	/* 
 	 * Check to see if any process groups have become orphaned
 	 * as a result of our exiting, and if they have any stopped
@@ -428,7 +428,7 @@ fake_volatile:
 		current->p_cptr = p->p_osptr;
 		p->p_ysptr = NULL;
 		p->flags &= ~(PF_PTRACED|PF_TRACESYS);
-		if (task[1])
+		if (task[1] && task[1] != current)
 			p->p_pptr = task[1];
 		else
 			p->p_pptr = task[0];
@@ -451,26 +451,13 @@ fake_volatile:
 			kill_pg(p->pgrp,SIGCONT,1);
 		}
 	}
-	if (current->leader) {
-		struct task_struct *p;
-		struct tty_struct *tty;
-
-		if (current->tty >= 0) {
-			tty = TTY_TABLE(current->tty);
-			if (tty) {
-				if (tty->pgrp > 0)
-					kill_pg(tty->pgrp, SIGHUP, 1);
-				tty->pgrp = -1;
-				tty->session = 0;
-			}
-		}
-		for_each_task(p) {
-			if (p->session == current->session)
-				p->tty = -1;
-		}
-	}
+	if (current->leader)
+		disassociate_ctty(1);
 	if (last_task_used_math == current)
 		last_task_used_math = NULL;
+	current->state = TASK_ZOMBIE;
+	current->exit_code = code;
+	current->rss = 0;
 #ifdef DEBUG_PROC_TREE
 	audit_ptree();
 #endif
